/*--------------------------------------------------------------------
  frequency1: An audio frequency meter using Timer1 counter.
			  Handles 1 Hz to 1 MHz with readout on 6 digit LED display.
 
  Processor: PIC16F684
  
  Discussion:
    Timer1 is used as a 16 bit counter with external clock, the
    clock being the input signal. By turning on the timer for
    exactly 1 second, the frequency can be directly taken from 
    the timer register word. That is, until the count exceeds 64K,
    at which point the timer counter overflows. The overflows are 
    counted in the timer1 interrupt handler, which becomes the  
    high-order word of the count. So we can count pretty high as
    far as software is concerned. The count is limited to
    9,999,999 in the ltoa module because the actual high bound is
    999999 Hz on a 6 digit display.
    
    Regarding calibration: The accuracy depends on timer0 measuring
    one second intervals. Since a crystal is not being used with
    this project, the clock is not accurate. It is easy to tweak
    timer0 by adjusting the initial count value (TIMER0_START_CNT), 
    which lengthens or shortens the time that constitutes the one
    second interval.
--------------------------------------------------------------------*/                
#include <system.h>
#include "frequency2.h"

#pragma CLOCK_FREQ      4000000

// Config Bits: 11 0000 1100 0100
// bit 13-12: unimplemented, set to 1
// bit 11:    FCMEN  0 = fail safe clock disabled
// bit 10:    IESO   0 = internal external switchover disabled
// bits 9-8:  BODEN  00 = Brown out detect disabled
// bit 7:     !CPD   1 = data memory code protection disabled
// bit 6:     !CP    1 = program memory code protection disabled
// bit 5:     MCLRE  0 = RA3/!MCLR pin function is RA3
// bit 4:     ^PWRTE 0 = PWRT enabled
// bit 3:     WDTE   0 = watchdog timer disabled
// bits 2-0:  FOSC   100 = INTOSC oscillator, I/O on RA4/OSC2/CLKOUT, I/O on RA5/OSC1/CLKIN

#pragma DATA _CONFIG, 0x30c4

#define RA4             0x0004		// diagnostic LED

#define TIMER0_START_CNT	51		// tewak this value to calibrate

// Global Variables

static unsigned short usOverflowCtr;
static unsigned long  ulTimer1Ctr;

// Functions

static void initCPU( void );
static void ledDigit( char cDigit );
static void ledOutput( char cNumber );

/*---------------------------------------------------------------------
  main()
---------------------------------------------------------------------*/
void main()
{
  char buffer[8];
  char cDigit, cNumber, cLen, cStrt;
  
  initCPU();                	// initialize hardware
  usOverflowCtr = 0;
  ulTimer1Ctr = 0;
  set_bit( t1con, TMR1ON );		// start timer1 counter

  // The mainline job is to convert timer1 counts to ascii and display
  // the LED digits. Everything else is managed by timer0 and timer1
  // interrupt handlers.

  while ( 1 )
  {
    // The number in buffer has no leading zeroes. We want to display
    // the digits right-justified. To do this we skip high-order
    // digits when there is nothing to display.
   
    ltoa( ulTimer1Ctr, buffer );
    cLen = strlen( buffer );
    cStrt = 6 - cLen;
    
    for ( cDigit=0; cDigit<cLen; ++cDigit, ++cStrt )
    {
      cNumber = buffer[cDigit];
      ledOutput( cNumber );				// output a number
      ledDigit( cStrt );				// turn the digit on
      delay_ms( 2 );
    }  
  }
}

/*---------------------------------------------------------------------
  interrupt()
  
  Handles the timer2 interrupt.
---------------------------------------------------------------------*/
void interrupt( void )
{
  static unsigned char t2ctr=0;
  static unsigned short t0ctr=TIMER0_START_CNT;
  unsigned short usTimer1_word;
  
  if ( test_bit( intcon, T0IF ) )
  {
    // The timer0 interrupt occurs ( Fosc/4 * prescale * t0 counts ). When 
    // Fosc/4 is 1 microsecond and prescale is 1:2, the timer increments
    // every 2 microseconds. If we count 250 times, we get an interrupt every 
    // 500 microseconds. Since we are timing one second intervals, we need to
    // count to 2000.
    ++t0ctr;
    if ( t0ctr == 2000 )
    {
      clear_bit( t1con, TMR1ON );		// stop timer1 counter
      MAKESHORT( usTimer1_word, tmr1l, tmr1h );
      ulTimer1Ctr = ( (unsigned long) usOverflowCtr << 16 ) | (unsigned long) usTimer1_word;

      usOverflowCtr = 0;
      tmr1l = tmr1h = 0;
      set_bit( t1con, TMR1ON );		// start timer1 counter
      t0ctr = TIMER0_START_CNT;		// reset the counter
    }
    
    clear_bit( intcon, T0IF );		// reset timer0 overflow interrupt flag
    return;
  }
  
  if ( test_bit( pir1, TMR1IF ) )
  {
    // The timer1 interrupt occurs when Timer1, operating as a 
    // counter, rolls over its 16 bit register. We just like to 
    // keep track of the overflows in a separate counter.
    ++usOverflowCtr;
    clear_bit( pir1, TMR1IF );		// reset timer1 overflow interrupt flag
    return;
  }

  if ( test_bit( pir1, TMR2IF ) )
  {
    // The timer2 interrupt occurs ( t2 counts * prescale * postscale * Fosc/4 ).
    // When Fosc/4 is 1 microsecond, and pre and postscale are 16,
    // the interrupt occurs every 65.5 milliseconds. Anyway all we do here
    // is turn the final decimal point on and off.

    ++t2ctr;
    if ( t2ctr == 8 )
      set_bit( porta, RA4 );		// turn LED on
    else
      if ( t2ctr == 16 )
      {
        clear_bit( porta, RA4 );
        t2ctr = 0;
      }
    
    clear_bit( pir1, TMR2IF );		// reset timer2 overflow interrupt flag
  }
}

/*---------------------------------------------------------------------
  initCPU()
---------------------------------------------------------------------*/
static void initCPU()
{
  // Initialize the internal clock:
  osccon = 01100000b;		// internal clock @ 4MHz, use FOSC (config)

  // PORTC connections as follows:
  //  RC0 - (out) LED data 'A'
  //  RC1 - (out) LED data 'B'
  //  RC2 - (out) LED data 'C'
  //  RC3 - (out) LED data 'D'
  //  RC4 - (out) LED digit select 'P1'
  //  RC5 - (out) LED digit select 'P2'

  portc = 0;
  cmcon0 = 00000111b;			// comparators off
  ansel = 0;					// digital I/O
  trisc = 00000000b;            // 1=input, 0=output

  // PORTA connections as follows:
  //  RA0/AN0/ICSPDAT - (in) (used for programming only)
  //  RA1/AN1/ICSPCLK - (in) (used for programming only)
  //  RA2/AN2/T0CKI - (out) LED digit select 'P0'
  //  RA3/!MCLR/Vpp - (in) (used for programming only)
  //  RA4/AN3 - (out) - diagnostic LED
  //  RA5/T1CKI - (in) signal input to timer1 counter

  porta = 00000000b;			// make RA4 low, LED off
  trisa = 00101011b;			// 1=input, 0=output

  // Timer0 initialization. Use Timer0 to measure one second intervals.
  tmr0 = 0;
  option_reg = 11010000b;		// internal clock; 1:2 prescale;
  set_bit( intcon, T0IE );		// enable timer0 interrupts
  
  // Timer1 initialization. Use Timer1 as 16-bit counter with
  // input on RA5/T1CKI.

  tmr1l = tmr1h = 0;
  t1con = 00000110b;      	// 1:1 prescale; LP osc off; no sync; ext. clock; stopped
  set_bit( pie1, TMR1IE );      // Timer1 interrupts

  // Timer2 initialization. Timer2 is used to generate interrupts
  // to turn an LED on and off, to show the program is running.
  
  t2con = 01111111b;			// 1:16 postscale; timer2 on; 1:16 prescale
  set_bit( pie1, TMR2IE );      // Timer2 interrupts

  // Enable interrupts for Timers

  set_bit( intcon, GIE );       // enable global interrupts
  set_bit( intcon, PEIE );      // enable peripheral interrupts
}

/*---------------------------------------------------------------------
  ledDigit(...)
  
  Selects a digit of the led display. Ports RA2-RC4-RC5 are connected
  to the A-B-C inputs of a 74LS145 (input D is grounded and not used).
  The '145 decodes the input and selects one of the 6 LED digits. Note
  that A is the low-order input and C is the high order input to the 
  '145.
  
  Input: a single numeric digit, range 0..5.
---------------------------------------------------------------------*/
static void ledDigit( char cDigit )
{
  if ( test_bit( cDigit, 0 ) )
    set_bit( porta, 2 );
  else
    clear_bit( porta, 2 );

  if ( test_bit( cDigit, 1 ) )
    set_bit( portc, 4 );
  else
    clear_bit( portc, 4 );

  if ( test_bit( cDigit, 2 ) )
    set_bit( portc, 5 );
  else
    clear_bit( portc, 5 );
}

/*---------------------------------------------------------------------
  ledOutput(...)
  
  Outputs a number to the led display. Ports RC0..RC3 are connected
  to the A-B-C-D inputs of a 74LS47 which decodes the number and
  activates the corresponding segments of the bank of LEDs. Note that
  A is the low-order input and D is the high order input to the '47.
  
  Input: a single numeric digit, range 0..9.
---------------------------------------------------------------------*/
static void ledOutput( char cNumber )
{
  // Output the digit as BCD on RC0..RC3

  if ( test_bit( cNumber, 0 ) )
    set_bit( portc, 0 );
  else
    clear_bit( portc, 0 );

  if ( test_bit( cNumber, 1 ) )
    set_bit( portc, 1 );
  else
    clear_bit( portc, 1 );

  if ( test_bit( cNumber, 2 ) )
    set_bit( portc, 2 );
  else
    clear_bit( portc, 2 );

  if ( test_bit( cNumber, 3 ) )
    set_bit( portc, 3 );
  else
    clear_bit( portc, 3 );
}
